"""
This module generates rivalcfg's CLI.
"""


import os
import sys
import types
import platform
import argparse

from . import handlers
from . import devices
from . import udev
from .version import VERSION


def normalize_cli_option_name(name):
    """Helper function to transform a setting name to a cli option.

    :param str name: The setting name.
    :rtype: str

    >>> normalize_cli_option_name("My_Test_Setting1")
    'my-test-setting1'
    """
    return name.lower().replace("_", "-")


class PrintSupportedDevicesAction(argparse.Action):
    """Print supported devices and exit."""

    def __call__(self, parser, namespace, value, option_string=None):
        for item in [getattr(devices, name) for name in dir(devices)]:
            if not isinstance(item, types.ModuleType):
                continue
            if not hasattr(item, "profile"):
                continue
            print("%s:" % item.profile["name"])
            print()
            for model in item.profile["models"]:
                print("  %04x:%04x | %s" % (
                    model["vendor_id"],
                    model["product_id"],
                    model["name"]))
            print()
        sys.exit(0)


class UpdateUdevRulesAction(argparse.Action):
    """Updates udev rules and exit."""

    def __call__(self, parser, namespace, value, option_string=None):
        if platform.system() != "Linux":
            print("E: The --update-udev option can only be used on Linux.")
            sys.exit(2)
        if os.getuid() != 0:
            print("E: You must run rivalcfg as root to use the --update-udev option.")  # noqa
            sys.exit(2)
        udev.write_rules_file()
        udev.trigger()
        sys.exit(0)


def add_main_cli(cli_parser):
    """Adds the main CLI options.

    :param ArgumentParser cli_parser: An :class:`ArgumentParser` instance.
    """
    cli_parser.add_argument(
            "--list",
            help="List supported devices and exit",
            nargs=0,
            action=PrintSupportedDevicesAction)

    cli_parser.add_argument(
            "--version",
            action="version",
            version=VERSION)

    cli_parser.add_argument(
            "--no-save",
            help="Do not persist settings in the internal device memory",
            dest="SAVE",
            action="store_false",
            default=True)

    cli_parser.add_argument(
            "--update-udev",
            help="Updates udev rules (Linux only, requires to be run as root)",
            nargs=0,
            action=UpdateUdevRulesAction)
    # TODO --print-info
    pass


def add_mouse_cli(cli_parser, mouse_profile):
    """Adds the CLI options for the given mouse profile.

    :param ArgumentParser cli_parser: An :class:`ArgumentParser` instance.
    :param mouse_profile: One of the rivalcfg mouse profile (provided by
                            :func:`rivalcfg.devices.get_profile`).
    """
    cli_group = cli_parser.add_argument_group(
            "%s Options" % mouse_profile["name"])

    for setting_name, setting_info in mouse_profile["settings"].items():
        handler = getattr(handlers, setting_info["value_type"])
        handler.add_cli_option(cli_group, setting_name, setting_info)

    cli_group.add_argument(
            "-r", "--reset",
            help="Reset all settings to their factory default",
            dest="RESET",
            action="store_true")



from rivalcfg import devices


class TestListPluggedDevice(object):

    def test_debug_devices(self, monkeypatch):
        monkeypatch.setenv("RIVALCFG_PROFILE", "1038:1702")
        devices_list = list(devices.list_plugged_devices())
        debug_device_found = False
        for device in devices_list:
            if device["vendor_id"] == 0x1038 \
               and device["product_id"] == 0x1702 \
               and device["name"] == "SteelSeries Rival 100":
                debug_device_found = True
        assert debug_device_found



import pytest

from rivalcfg import udev


class TestCheckUdevRules(object):

    @pytest.fixture
    def rules(self):
        return "# Generated by rivalcfg v42.0.0\nFoo\nBar...\n"

    def test_right_version(self, rules):
        assert udev.are_rules_up_to_date(rules, current_version="42.0.0") is True  # noqa

    def test_wrong_version(self, rules):
        assert udev.are_rules_up_to_date(rules, current_version="13.37.0") is False  # noqa


class TestReadWriteUdevRulesFile(object):

    def test_write_and_read_file(self, tmp_path):
        path = tmp_path / "test.rules"
        udev.write_rules_file(path=path)
        assert udev.is_rules_file_up_to_date(path=path) is True



from .. import usbhid


profile = {

    "name": "SteelSeries Kinzu v2",

    "models": [{
        "name": "SteelSeries Kinzu v2",
        "vendor_id": 0x1038,
        "product_id": 0x1366,
        "endpoint": 0,
    }, {
        "name": "SteelSeries Kinzu v2",
        "vendor_id": 0x1038,
        "product_id": 0x1378,
        "endpoint": 0,
    }],

    "settings": {

        "sensitivity1": {
            "label": "Sensibility preset 1",
            "description": "Set sensitivity preset 1 (DPI)",
            "cli": ["-s", "--sensitivity1"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x03, 0x01],
            "value_type": "choice",
            "choices": {
                400: 0x04,
                800: 0x03,
                1600: 0x02,
                3200: 0x01,
            },
            "default": 800,
        },

        "sensitivity2": {
            "label": "Sensibility preset 2",
            "description": "Set sensitivity preset 2 (DPI)",
            "cli": ["-S", "--sensitivity2"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x03, 0x02],
            "value_type": "choice",
            "choices": {
                400: 0x04,
                800: 0x03,
                1600: 0x02,
                3200: 0x01,
            },
            "default": 3200,
        },

        "polling_rate": {
            "label": "Polling rate",
            "description": "Set polling rate (Hz)",
            "cli": ["-p", "--polling-rate"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x04, 0x00],
            "value_type": "choice",
            "choices": {
                125: 0x04,
                250: 0x03,
                500: 0x02,
                1000: 0x01,
            },
            "default": 1000,
        },

    },

    "save_command": {
        "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
        "command": [0x09, 0x00],
    },

}



"""
This module contains low level functions to interact with USB HID devices. The
`hidapi <https://pypi.org/project/hidapi/>`_ module is used to abstract the
access to the devices across all different operating systems.

This module can also simulate devices:

* if you define the ``RIVALCFG_DRY`` environment variable, a fake device will
  be returned instead of a real one. This is useful for debugging and testing::

    RIVALCFG_DRY=1 rivalcfg -c ff0000

* If you define the ``RIVALCFG_PROFILE`` environment variable with a vendor id
  and a product id, this module will report the corresponding device as being
  plugged::

    RIVALCFG_PROFILE=1038:1702 rivalcfg -h
"""


import os
import struct
from io import BytesIO

import hid


#: HID output report
HID_REPORT_TYPE_OUTPUT = 0x02

#: HID feature report
HID_REPORT_TYPE_FEATURE = 0x03


def is_device_plugged(vendor_id, product_id):
    """Returns ``True`` if the given HID device is plugged to the computer.

    :param int vendor_id: The vendor id of the device (e.g. ``0x1038``)).
    :param int product_id: The product id of the device (e.g. ``0x1710``).

    :rtype: bool

    >>> from rivalcfg import usbhid
    >>> usbhid.is_device_plugged(0x1038, 0xbaad)
    False
    """
    if "RIVALCFG_PROFILE" in os.environ:
        debug_vendor_id = int(os.environ["RIVALCFG_PROFILE"].split(":")[0], 16)
        debug_product_id = int(
                os.environ["RIVALCFG_PROFILE"].split(":")[1], 16)
        if debug_vendor_id == vendor_id and debug_product_id == product_id:
            return True
    return len(hid.enumerate(vendor_id, product_id)) > 0


def open_device(vendor_id, product_id, endpoint):
    """Opens and returns the HID device

    :param int vendor_id: The vendor id of the device (e.g. ``0x1038``)).
    :param int product_id: The product id of the device (e.g. ``0x1710``).
    :param int endpoint: The number of the endpoint to open on the device (e.g.
                         ``0``).

    :raise DeviceNotFound: The requested device is not plugged to the computer
                           or it does not provide the requested endpoint.
    :raise IOError: The device or its interface cannot be opened (permission
                    issue, device busy,...).

    :rtype: hid.device

    >>> from rivalcfg import usbhid
    >>> usbhid.open_device(0x1038, 0x1702, 0)
    <hid.device at 0x...>
    """
    # Instanciate the device (real or fake depending of environment)
    if "RIVALCFG_DRY" in os.environ:
        # Setting a path allows the fake device to be opened even if requested
        # device is not plugged to the computer
        path = b"00:0000:0000"
        device = FakeDevice()
    else:
        path = None
        device = hid.device()

    # Search the device
    for interface in hid.enumerate(vendor_id, product_id):
        if interface["interface_number"] == endpoint:
            path = interface["path"]
            break

    # Open the found device. This can raise an IOError.
    if path:
        device.open_path(path)
        return device

    # No matching device found
    raise DeviceNotFound("Requested device or endpoint not found: %04x:%04x:%02x" % (  # noqa
        vendor_id, product_id, endpoint))


class DeviceNotFound(Exception):
    """Exception raised when the requested device was not found (device not
    plugged to the computer) or when it does not provide the requested
    endpoint.
    """


class FakeDevice:

    """This class simulate an HID device as provided by the `hidapi
    <https://pypi.org/project/hidapi/>`_ module.

    Data sent to the fake device can be read-back from the :attr:`bytes`
    attribute of this class.
    """

    def __init__(self):
        """Constructor."""

        #: A BytesIO instance where every simulated write will be done.
        self.bytes = BytesIO()

    def open_path(self, path):
        """Simulate opening the device from given path.

        :param bytes path: The device path.
        """
        pass

    def write(self, data):
        """Simulate a write to an HID device.

        Given data will be writen to :attr:`bytes`, prefixed with the
        :data:`HID_REPORT_TYPE_OUTPUT` byte.

        :param bytes data: The data to send to the device.

        >>> from rivalcfg.usbhid import FakeDevice
        >>> dev = FakeDevice()
        >>> dev.write(b"\\x00\\xAA\\xBB\\xCC")
        >>> dev.bytes.seek(0)
        0
        >>> dev.bytes.read()
        b'\\x02\\x00\\xaa\\xbb\\xcc'
        """
        self.bytes.write(struct.pack("B", HID_REPORT_TYPE_OUTPUT))
        self.bytes.write(data)

    def send_feature_report(self, data):
        """Simulate a feature report write to an HID device.

        Given data will be writen to :attr:`bytes`, prefixed with the
        :data:`HID_REPORT_TYPE_FEATURE` byte.

        :param bytes data: The data to send to the device.

        >>> from rivalcfg.usbhid import FakeDevice
        >>> dev = FakeDevice()
        >>> dev.send_feature_report(b"\\x00\\xAA\\xBB\\xCC")
        >>> dev.bytes.seek(0)
        0
        >>> dev.bytes.read()
        b'\\x03\\x00\\xaa\\xbb\\xcc'
        """
        self.bytes.write(struct.pack("B", HID_REPORT_TYPE_FEATURE))
        self.bytes.write(data)

    def close(self):
        """Closes the simulated device."""
        self.bytes.close()

    def error(self):
        raise NotImplementedError()

    def get_feature_report(self):
        raise NotImplementedError()

    def get_manufacturer_string(self):
        raise NotImplementedError()

    def get_product_string(self):
        raise NotImplementedError()

    def get_serial_number_string(self):
        raise NotImplementedError()

    def open(self):
        raise NotImplementedError()

    def read(self):
        raise NotImplementedError()

    def set_nonblocking(self):
        raise NotImplementedError()

    def __repr__(self):
        return "<hid.device at 0xbadc0fee%x>" % id(self)



import os
import sys
import platform
import argparse

from . import cli
from . import udev
from . import devices
from . import get_first_mouse


_EPILOG = """
Please report any bug on Github: https://github.com/flozz/rivalcfg/issues
"""


def _check_linux():
    """Run some checks on Linux."""

    # Only run checks on Linux
    if platform.system() != "Linux":
        return

    # Skip check for root as there will be no issue for this user
    if os.getuid() == 0:
        return

    # Skip check if user is trying to update udev rules:
    if "--update-udev" in sys.argv:
        return

    if not os.path.isfile(udev.RULES_FILE_PATH):
        print("W: udev rules are not installed. You may not be able to open the device using a regular user.")  # noqa
        print("\n   Run 'rivalcfg --update-udev' as root to fix.\n")
    elif not udev.is_rules_file_up_to_date():
        print("W: Installed udev rules were generated by an other rivalcfg version.")  # noqa
        print("\n   Run 'rivalcfg --update-udev' as root to update.\n")


def main(args=sys.argv[1:]):
    # On Linux: check udev rules
    _check_linux()

    # Try to open a mouse
    mouse = get_first_mouse()

    cli_parser = argparse.ArgumentParser(
            prog="rivalcfg",
            epilog=_EPILOG)
    cli.add_main_cli(cli_parser)

    if mouse:
        mouse_profile = devices.get_profile(
                vendor_id=mouse.vendor_id,
                product_id=mouse.product_id)
        cli.add_mouse_cli(cli_parser, mouse_profile)

    settings = cli_parser.parse_args(args)

    # Reset
    if mouse and settings.RESET:
        mouse.reset_settings()

    # Apply settings
    if mouse:
        for setting_name, value in [
                (k.lower(), v) for k, v in vars(settings).items()]:
            if value is None:
                continue
            method_name = "set_%s" % setting_name
            if not hasattr(mouse, method_name):
                continue
            getattr(mouse, method_name)(value)

        # Save settings in the internal device memory
        if settings.SAVE:
            mouse.save()


if __name__ == "__main__":
    main()



"""
This file contains the layout for QWERTY (en_US) keyboards.
"""


#: The layout
layout = {

    # Alphanumeric

    "A": 0x04,
    "B": 0x05,
    "C": 0x06,
    "D": 0x07,
    "E": 0x08,
    "F": 0x09,
    "G": 0x0A,
    "H": 0x0B,
    "I": 0x0C,
    "J": 0x0D,
    "K": 0x0E,
    "L": 0x0F,
    "M": 0x10,
    "N": 0x11,
    "O": 0x12,
    "P": 0x13,
    "Q": 0x14,
    "R": 0x15,
    "S": 0x16,
    "T": 0x17,
    "U": 0x18,
    "V": 0x19,
    "W": 0x1A,
    "X": 0x1B,
    "Y": 0x1C,
    "Z": 0x1D,

    "1": 0x1E,
    "2": 0x1F,
    "3": 0x20,
    "4": 0x21,
    "5": 0x22,
    "6": 0x23,
    "7": 0x24,
    "8": 0x25,
    "9": 0x26,
    "0": 0x27,

    # Editing

    "Enter":     0x28,
    "Escape":    0x29,
    "BackSpace": 0x2A,
    "Tab":       0x2B,
    "Space":     0x2C,
    "Delete":    0x4C,

    # Symbols

    "-": 0x2D,
    "=": 0x2E,
    "[": 0x2F,
    "]": 0x30,
    "\\": 0x31,
    "#": 0x32,  # International!?
    ";": 0x33,
    "'": 0x34,
    "`": 0x35,
    ",": 0x36,
    ".": 0x37,
    "/": 0x38,
    "\\(inter)": 0x64,  # International!?

    # Typing Mode

    "CapsLock":    0x39,
    "ScrollLock":  0x47,
    "Insert":      0x49,
    "NumLock":     0x53,

    # Functions

    "F1":  0x3A,
    "F2":  0x3B,
    "F3":  0x3C,
    "F4":  0x3D,
    "F5":  0x3E,
    "F6":  0x3F,
    "F7":  0x40,
    "F8":  0x41,
    "F9":  0x42,
    "F10": 0x43,
    "F11": 0x44,
    "F12": 0x45,
    "F13": 0x68,
    "F14": 0x69,
    "F15": 0x6A,
    "F16": 0x6B,
    "F17": 0x6C,
    "F18": 0x6D,
    "F19": 0x6E,
    "F20": 0x6F,
    "F21": 0x70,
    "F22": 0x71,
    "F23": 0x72,
    "F24": 0x73,

    # Commands

    "PrintScreen": 0x46,
    "PauseBreak":  0x48,
    "ContextMenu": 0x65,

    # Navigation

    "Home":     0x4A,
    "PageUp":   0x4B,
    "End":      0x4D,
    "PageDown": 0x4E,

    # Arrows

    "Right": 0x4F,
    "Left":  0x50,
    "Down":  0x51,
    "Up":    0x52,

    # Numpad

    "Keypad/": 0x54,
    "Keypad*": 0x55,
    "Keypad-": 0x56,
    "Keypad+": 0x57,
    "KeypadEnter": 0x58,
    "Keypad1": 0x59,
    "Keypad2": 0x5A,
    "Keypad3": 0x5B,
    "Keypad4": 0x5C,
    "Keypad5": 0x5D,
    "Keypad6": 0x5E,
    "Keypad7": 0x5F,
    "Keypad8": 0x60,
    "Keypad9": 0x61,
    "Keypad0": 0x62,
    "Keypad.": 0x63,
    "Keypad,": 0x85,  # ??
    "Keypad=": 0x86,  # ??

    # Modifiers

    "LeftCtrl":   0xE0,
    "LeftShift":  0xE1,
    "LeftAlt":    0xE2,
    "LeftSuper":  0xE3,  # Command / Win logo
    "RightCtrl":  0xE4,
    "RightShift": 0xE5,
    "RightAlt":   0xE6,
    "RightSuper": 0xE7,  # Command / Win logo

}

#: Alias for some keys of the layout
aliases = {
    "esc": "Escape",
    "bksp": "BackSpace",
    "bkspace": "BackSpace",
    "del": "Delete",

    "dash": "-",
    "minus": "-",
    "equal": "=",
    "eq": "=",
    "leftbracket": "[",
    "rightbracket": "]",
    "backslash": "\\",
    "hash": "#",
    "semicolon": ";",
    "semi": ";",
    "quote": "'",
    "backtick": "`",
    "backquote": "`",
    "comma": ",",
    "dot": ".",
    "point": ".",
    "slash": "/",

    "capslck": "CapsLock",
    "capslk": "CapsLock",
    "cpslck": "CapsLock",
    "cpslk": "CapsLock",
    "scrolllck": "ScrollLock",
    "scrolllk": "ScrollLock",
    "scrllck": "ScrollLock",
    "scrllk": "ScrollLock",
    "scrlck": "ScrollLock",
    "scrlk": "ScrollLock",
    "ins": "Insert",
    "num": "NumLock",
    "numlck": "NumLock",
    "numlk": "NumLock",

    "prntscr": "PrintScreen",
    "prtscr": "PrintScreen",
    "prtsc": "PrintScreen",
    "psbrk": "PauseBreak",
    "psbr": "PauseBreak",
    "ctx": "ContextMenu",
    "menu": "ContextMenu",
    "ctxmenu": "ContextMenu",
    "ctxmn": "ContextMenu",

    "pgup": "PageUp",
    "pgdown": "PageDown",
    "pgdwn": "PageDown",
    "pgdn": "PageDown",

    "lctrl": "LeftCtrl",
    "lshift": "LeftShift",
    "lalt": "LeftAlt",
    "alt": "LeftAlt",
    "super": "LeftSuper",
    "lsuper": "LeftSuper",
    "windows": "LeftSuper",
    "leftwindows": "LeftSuper",
    "win": "LeftSuper",
    "lwin": "LeftSuper",
    "command": "LeftSuper",
    "leftcommand": "LeftSuper",
    "cmd": "LeftSuper",
    "lcmd": "LeftSuper",
    "rctrl": "RightCtrl",
    "rshift": "RightShift",
    "ralt": "RightAlt",
    "altgr": "RightAlt",
    "rsuper": "RightSuper",
    "rightwindows": "RightSuper",
    "rwin": "RightSuper",
    "rightcommand": "RightSuper",
    "rcmd": "RightSuper",
}



import os

from . import devices
from . import mouse


def get_first_mouse():
    """Get the first plugged mouse that can befound. As there is only one mouse
    plugged to a computer most of the time, it should return the right one. ;)

    :rtype: rivalcfg.mouse.Mouse
    :return: The mouse if one supported device is found, else returns ``None``.

    Always return ``None`` when ``RIVALCFG_PROFILE=0000:0000`` is defined in
    the environment.

    >>> import rivalcfg
    >>> rivalcfg.get_first_mouse()  # doctest: +SKIP
    <Mouse ...>
    """
    if "RIVALCFG_PROFILE" in os.environ \
            and os.environ["RIVALCFG_PROFILE"] == "0000:0000":
        return None
    plugged_devices = list(devices.list_plugged_devices())
    if not plugged_devices:
        return None
    return mouse.get_mouse(
            vendor_id=plugged_devices[0]["vendor_id"],
            product_id=plugged_devices[0]["product_id"])



"""
The "rgbcolor" type handles RGB color values.

It supports hexadecimal colors:

* ``#FF0000``
* ``FF0000``
* ``#F00``
* ``F00``

RGB tuples or list (only from Python API for this one):

* ``(255, 0, 0)``
* ``[255, 0, 0]``

and named colors:

+------------+------------+-----------+-------------+
| ``white``  | ``red``    | ``lime``  | ``blue``    |
+------------+------------+-----------+-------------+
| ``silver`` | ``maroon`` | ``green`` | ``navy``    |
+------------+------------+-----------+-------------+
| ``gray``   | ``yellow`` | ``aqua``  | ``fuchsia`` |
+------------+------------+-----------+-------------+
| ``black``  | ``olive``  | ``teal``  | ``purple``  |
+------------+------------+-----------+-------------+


Device Profile
--------------

Example of a rgbcolor value type in a device profile:

::

    profile = {

        # ...

        "settings": {

            "color": {
                "label": "LED color",
                "description": "Set the mouse backlight color",
                "cli": ["-c", "--color"],
                "command": [0x05, 0x00],
                "value_type": "rgbcolor",
                "default": "#FF1800"
            },

        },

        # ...

    }


CLI
---

Example of CLI option generated with this handler::

    -c COLOR, --color=COLOR
                        Set the mouse backlight color (e.g. red, #ff0000,
                        ff0000, #f00, f00, default: #FF1800)

Example of CLI usage::

    rivalcfg --color=red
    rivalcfg --color=ff0000
    rivalcfg --color=f00


Functions
---------
"""


import argparse

from ..color_helpers import is_color, parse_color_string


def process_value(setting_info, color):
    """Called by the :class:`rivalcfg.mouse.Mouse` class when processing a
    "rgbcolor" type setting.

    :param dict setting_info: The information dict of the setting from the
                              device profile.
    :param str,tuple,list color: The color.
    :rtype: [int, int, int]
    """
    # Color tuple
    if type(color) in (tuple, list):
        if len(color) != 3:
            raise ValueError("Not a valid color %s" % str(color))
        for channel in color:
            if type(channel) != int or channel < 0 or channel > 255:
                raise ValueError("Not a valid color %s" % str(color))
        return list(color)

    if is_color(color):
        return list(parse_color_string(color))

    raise ValueError("Not a valid color %s" % str(color))


class CheckColorAction(argparse.Action):
    """Validate colors from CLI"""

    def __call__(self, parser, namespace, value, option_string=None):
        if not is_color(value):
            raise argparse.ArgumentError(self, "invalid color: '%s'" %  value)  # noqa
        setattr(namespace, self.dest.upper(), value)


def add_cli_option(cli_parser, setting_name, setting_info):
    """Add the given "rgbcolor" type setting to the given CLI arguments parser.

    :param ArgumentParser cli_parser: An :class:`ArgumentParser` instance.
    :param str setting_name: The name of the setting.
    :param dict setting_info: The information dict of the setting from the
                              device profile.
    """
    description = "%s (e.g. red, #ff0000, ff0000, #f00, f00, default: %s)" % (
            setting_info["description"],
            str(setting_info["default"])
            )
    cli_parser.add_argument(
            *setting_info["cli"],
            dest=setting_name,
            help=description,
            type=str,
            action=CheckColorAction,
            metavar=setting_name.upper()
            )



from .. import usbhid


_RGBGRADIENT_HEADER = {
    "header_length": 26,       # Length of the header excuding command / LED ID
    "led_id_offsets": [0],     # Offset of the "led_id" fields
    "duration_offset": 1,      # Offset of the "duration" field
    "duration_length": 2,      # Length of the "duration" field (in Bytes)
    "repeat_offset": 17,       # Offset of the "repeat" flag
    "triggers_offset": 21,     # Offset of the "triggers" field (buttons mask)
    "color_count_offset": 25,  # Offset of the "color_count" field
}

_DEFAULT_RGBGRADIENT = "rgbgradient(duration=1000; colors=0%: #ff0000, 33%: #00ff00, 66%: #0000ff)"  # noqa


profile = {

    "name": "SteelSeries Sensei TEN",

    "models": [{
        "name": "SteelSeries Sensei TEN",
        "vendor_id": 0x1038,
        "product_id": 0x1832,
        "endpoint": 0,
    }, {
        "name": "SteelSeries Sensei TEN CS:GO Neon Rider Editon",
        "vendor_id": 0x1038,
        "product_id": 0x1834,
        "endpoint": 0,
    }],

    "settings": {

        "sensitivity": {
            "label": "Sensibility presets",
            "description": "Set sensitivity preset (DPI)",
            "cli": ["-s", "--sensitivity"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x55, 0x00],  # <COUNT> <SELECTED> <S1> <S2> ... <S5>
            "value_type": "multidpi_range",
            "input_range": [50, 18000, 50],
            "output_range": [1, 0x0168, 1],
            "dpi_length_byte": 2,
            "count_mode": "flag",
            "max_preset_count": 5,
            "default": "800, 1600",
        },

        "polling_rate": {
            "label": "Polling rate",
            "description": "Set polling rate (Hz)",
            "cli": ["-p", "--polling-rate"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x54, 0x00],
            "value_type": "choice",
            "choices": {
                125: 0x04,
                250: 0x03,
                500: 0x02,
                1000: 0x01,
            },
            "default": 1000,
        },

        "logo_color": {
            "label": "Logo LED colors and effects",
            "description": "Set the colors and the effects of the logo LED",
            "cli": ["-c", "--logo-color"],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "command": [0x5B, 0x00],
            "value_type": "rgbgradient",
            "rgbgradient_header": _RGBGRADIENT_HEADER,
            "led_id": 0x00,
            "default": _DEFAULT_RGBGRADIENT,
        },

        "wheel_color": {
            "label": "Wheel LED colors and effects",
            "description": "Set the colors and the effects of the wheel LED",
            "cli": ["-C", "--wheel-color"],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "command": [0x5B, 0x00],
            "value_type": "rgbgradient",
            "rgbgradient_header": _RGBGRADIENT_HEADER,
            "led_id": 0x01,
            "default": _DEFAULT_RGBGRADIENT,
        },

        "buttons_mapping": {
            "label": "Buttons mapping",
            "description": "Set the mapping of the buttons",
            "cli": ["-b", "--buttons"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x31, 0x00],
            "value_type": "buttons",

            "buttons": {
                "Button1": {"id": 0x01, "offset": 0x00, "default": "button1"},
                "Button2": {"id": 0x02, "offset": 0x05, "default": "button2"},
                "Button3": {"id": 0x03, "offset": 0x0A, "default": "button3"},
                "Button4": {"id": 0x04, "offset": 0x0F, "default": "button4"},
                "Button5": {"id": 0x05, "offset": 0x14, "default": "button5"},
                "Button6": {"id": 0x06, "offset": 0x19, "default": "PageDown"},
                "Button7": {"id": 0x07, "offset": 0x1E, "default": "PageUp"},
                "Button8": {"id": 0x08, "offset": 0x23, "default": "dpi"},
            },

            "button_field_length": 5,

            "button_disable":     0x00,
            "button_keyboard":    0x51,
            "button_multimedia":  0x61,
            "button_dpi_switch":  0x30,
            "button_scroll_up":   0x31,
            "button_scroll_down": 0x32,

            "default": "buttons(button1=button1; button2=button2; button3=button3; button4=button4; button5=button5; button6=PageDown; button7=PageUp; button8=dpi; layout=qwerty)",  # noqa
        },

    },

    "save_command": {
        "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
        "command": [0x59, 0x00],
    },

}



import time

from . import usbhid
from . import devices
from . import handlers
from . import helpers


def get_mouse(vendor_id=0x1038, product_id=None):
    """Get a :class:`Mouse` instance to manipulate requested device.

    :param int vendor_id: The vendor id of the device (optional, by default
                          this is set to the SteelSeries vendor id
                          (``0x1038``)).
    :param int product_id: The product id of one of the supported device (e.g.
                           ``0x1702``).
    :raise rivalcfg.devices.UnsupportedDevice: The requested device is not
                                               supported by rivalcfg.
    :rtype: Mouse

    >>> from rivalcfg.mouse import get_mouse
    >>> get_mouse(vendor_id=0x1038, product_id=0x1702)
    <Mouse SteelSeries Rival 100 (1038:1702:00)>
    """
    if not product_id:
        raise ValueError("You must define the 'product_id' parameter")

    profile = devices.get_profile(vendor_id, product_id)

    hid_device = usbhid.open_device(vendor_id, product_id, profile["endpoint"])

    return Mouse(hid_device, profile)


class Mouse:

    """Generic class to handle any supported mouse.

    .. NOTE::

       Additional methods are available in this class depending on the loaded
       profile. Read device specific documentation for more information.

    .. WARNING::

       You should not instanciate this class yourself. Use the
       :func:`get_mouse` factory function instead.

    :param hid_device: The HID device to write in (provided by the
                        :func:`rivalcfg.usbhid.open_device`).
    :param mouse_profile: One of the rivalcfg mouse profile (provided by
                            :func:`rivalcfg.devices.get_profile`).

    >>> from rivalcfg import usbhid
    >>> from rivalcfg import devices
    >>> from rivalcfg.mouse import Mouse
    >>> Mouse(
    ...     usbhid.open_device(vendor_id=0x1038, product_id=0x1702, endpoint=0),
    ...     devices.get_profile(vendor_id=0x1038, product_id=0x1702))
    <Mouse SteelSeries Rival 100 (1038:1702:00)>
    """  # noqa

    def __init__(self, hid_device, mouse_profile):
        """Constructor."""
        self._hid_device = hid_device
        self._mouse_profile = mouse_profile

    @property
    def name(self):
        """The mouse name."""
        return self._mouse_profile["name"]

    @property
    def vendor_id(self):
        """The mouse vendor id."""
        return self._mouse_profile["vendor_id"]

    @property
    def product_id(self):
        """The mouse product id."""
        return self._mouse_profile["product_id"]

    def reset_settings(self):
        """Sets all settings to their factory default values."""
        for name, setting_info in self._mouse_profile["settings"].items():
            method_name = "set_%s" % name
            method = getattr(self, method_name)
            if "value_type" in setting_info and setting_info["value_type"]:
                method(setting_info["default"])
            else:
                method()

    def save(self):
        """Save current config to the mouse internal memory."""
        # This should never apand... But who knows...
        if "save_command" not in self._mouse_profile \
           or not self._mouse_profile["save_command"]:
            raise Exception("This mouse does not provide any save command.")

        packet_length = 0
        if "packet_length" in self._mouse_profile["save_command"]:
            packet_length = self._mouse_profile["save_command"]["packet_length"]  # noqa: E501

        self._hid_write(
                report_type=self._mouse_profile["save_command"]["report_type"],
                data=self._mouse_profile["save_command"]["command"],
                packet_length=packet_length)

    def _hid_write(self,
                   report_type=usbhid.HID_REPORT_TYPE_OUTPUT,
                   report_id=0x00,
                   data=[],
                   packet_length=0):
        """
        Write data to the device.

        :param int report_type: The HID report type
                                (:data:`rivalcfg.usbhid.HID_REPORT_TYPE_OUTPUT`
                                or
                                :data:`rivalcfg.usbhid.HID_REPORT_TYPE_FEATURE`).
        :param int report_id: The id of the report (always ``0x00``).
        :param list(int) data: The data to send to the mouse.
        :param int packet_length: The fixed length of the packet that will be
                                  sent to the device (default: ``0`` (no fixed
                                  length)).

        :raises ValueError: Invalid report type, or HID device not openned.
        """
        if packet_length:
            bytes_ = bytearray(helpers.merge_bytes(
                report_id,
                data,
                [0x00] * (packet_length - len(data))
                ))
        else:
            bytes_ = bytearray(helpers.merge_bytes(report_id, data))
        if report_type == usbhid.HID_REPORT_TYPE_OUTPUT:
            self._hid_device.write(bytes_)
        elif report_type == usbhid.HID_REPORT_TYPE_FEATURE:
            self._hid_device.send_feature_report(bytes_)
        else:
            raise ValueError("Invalid HID report type: %2x" % report_type)

        # Avoids sending multiple commands to quickly
        time.sleep(0.01)

    def __getattr__(self, name):
        # Handle every set_xxx methods generated from device's profiles

        if not name.startswith("set_"):
            raise AttributeError("Mouse instance has no attribute '%s'" % name)

        setting_name = name[4:]

        if setting_name not in self._mouse_profile["settings"]:
            raise AttributeError("Mouse instance has no attribute '%s'" % name)

        setting_info = self._mouse_profile["settings"][setting_name]

        handler_name = None

        if "value_type" in setting_info and setting_info["value_type"]:
            handler_name = setting_info["value_type"]
            if handler_name not in helpers.module_ls(handlers):
                raise ValueError("Unknown handler '%s' for '%s' setting of the %s" % (  # noqa
                    handler_name,
                    setting_name,
                    self._mouse_profile["name"]))

        packet_length = 0
        if "packet_length" in setting_info:
            packet_length = setting_info["packet_length"]

        suffix = []
        if "command_suffix" in setting_info:
            suffix = setting_info["command_suffix"]

        def _exec_command(*args):
            data = []
            if handler_name:
                data = getattr(handlers, handler_name) \
                        .process_value(setting_info, *args)
            self._hid_write(
                    report_type=setting_info["report_type"],
                    data=helpers.merge_bytes(
                        setting_info["command"],
                        data,
                        suffix
                        ),
                    packet_length=packet_length)

        return _exec_command

    def __repr__(self):
        return "<Mouse %s (%04x:%04x:%02x)>" % (
                self._mouse_profile["name"],
                self._mouse_profile["vendor_id"],
                self._mouse_profile["product_id"],
                self._mouse_profile["endpoint"])



from .. import usbhid


_RGBGRADIENT_HEADER = {
    "header_length": 28,       # Length of the header excuding command / LED ID
    "led_id_offsets": [0, 5],  # Offset of the "led_id" fields
    "duration_offset": 6,      # Offset of the "duration" field
    "duration_length": 2,      # Length of the "duration" field (in Bytes)
    "repeat_offset": 22,       # Offset of the "repeat" flag
    "triggers_offset": 23,     # Offset of the "triggers" field (buttons mask)
    "color_count_offset": 27,  # Offset of the "color_count" field
}

_DEFAULT_RGBGRADIENT = "rgbgradient(duration=1000; colors=0%: #ff0000, 33%: #00ff00, 66%: #0000ff)"  # noqa


profile = {

    "name": "SteelSeries Rival 600",

    "models": [{
        "name": "SteelSeries Rival 600",
        "vendor_id": 0x1038,
        "product_id": 0x1724,
        "endpoint": 0,
    }, {
        "name": "SteelSeries Rival 600 Dota 2 Edition",
        "vendor_id": 0x1038,
        "product_id": 0x172e,
        "endpoint": 0,
    }],

    "settings": {

        "sensitivity1": {
            "label": "Sensibility preset 1",
            "description": "Set sensitivity preset 1 (DPI)",
            "cli": ["-s", "--sensitivity1"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x03, 0x00, 0x01],
            "command_suffix": [0x00, 0x42],
            "value_type": "range",
            "input_range": [100, 12000, 100],
            "output_range": [0x00, 0x77, 1],
            "default": 800,
        },

        "sensitivity2": {
            "label": "Sensibility preset 2",
            "description": "Set sensitivity preset 2 (DPI)",
            "cli": ["-S", "--sensitivity2"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x03, 0x00, 0x02],
            "command_suffix": [0x00, 0x42],
            "value_type": "range",
            "input_range": [100, 12000, 100],
            "output_range": [0x00, 0x77, 1],
            "default": 1600,
        },

        "polling_rate": {
            "label": "Polling rate",
            "description": "Set polling rate (Hz)",
            "cli": ["-p", "--polling-rate"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x04, 0x00],
            "value_type": "choice",
            "choices": {
                125: 0x04,
                250: 0x03,
                500: 0x02,
                1000: 0x01,
            },
            "default": 1000,
        },

        "wheel_color": {
            "label": "Wheel LED colors and effects",
            "description": "Set the colors and the effects of the wheel LED",
            "cli": ["-C", "--wheel-color", "--z0"],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "command": [0x05, 0x00],
            "value_type": "rgbgradient",
            "rgbgradient_header": _RGBGRADIENT_HEADER,
            "led_id": 0x00,
            "default": _DEFAULT_RGBGRADIENT,
        },

        "logo_color": {
            "label": "Logo LED colors and effects",
            "description": "Set the colors and the effects of the logo LED",
            "cli": ["-c", "--logo-color", "--z1"],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "command": [0x05, 0x00],
            "value_type": "rgbgradient",
            "rgbgradient_header": _RGBGRADIENT_HEADER,
            "led_id": 0x01,
            "default": _DEFAULT_RGBGRADIENT,
        },

        "z2_color": {
            "label": "Left strip, top LED colors and effects",
            "description": "Set the colors and the effects of the top LED of the left strip",  # noqa
            "cli": ["--left-strip-top-color", "--z2"],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "command": [0x05, 0x00],
            "value_type": "rgbgradient",
            "rgbgradient_header": _RGBGRADIENT_HEADER,
            "led_id": 0x02,
            "default": _DEFAULT_RGBGRADIENT,
        },

        "z3_color": {
            "label": "Right strip, top LED colors and effects",
            "description": "Set the colors and the effects of the top LED of the right strip",  # noqa
            "cli": ["--right-strip-top-color", "--z3"],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "command": [0x05, 0x00],
            "value_type": "rgbgradient",
            "rgbgradient_header": _RGBGRADIENT_HEADER,
            "led_id": 0x03,
            "default": _DEFAULT_RGBGRADIENT,
        },

        "z4_color": {
            "label": "Left strip, middle LED colors and effects",
            "description": "Set the colors and the effects of the middle LED of the left strip",  # noqa
            "cli": ["--left-strip-middle-color", "--z4"],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "command": [0x05, 0x00],
            "value_type": "rgbgradient",
            "rgbgradient_header": _RGBGRADIENT_HEADER,
            "led_id": 0x04,
            "default": _DEFAULT_RGBGRADIENT,
        },

        "z5_color": {
            "label": "Right strip, middle LED colors and effects",
            "description": "Set the colors and the effects of the middle LED of the right strip",  # noqa
            "cli": ["--right-strip-middle-color", "--z5"],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "command": [0x05, 0x00],
            "value_type": "rgbgradient",
            "rgbgradient_header": _RGBGRADIENT_HEADER,
            "led_id": 0x05,
            "default": _DEFAULT_RGBGRADIENT,
        },

        "z6_color": {
            "label": "Left strip, bottom LED colors and effects",
            "description": "Set the colors and the effects of the bottom LED of the left strip",  # noqa
            "cli": ["--left-strip-bottom-color", "--z6"],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "command": [0x05, 0x00],
            "value_type": "rgbgradient",
            "rgbgradient_header": _RGBGRADIENT_HEADER,
            "led_id": 0x06,
            "default": _DEFAULT_RGBGRADIENT,
        },

        "z7_color": {
            "label": "Right strip, bottom LED colors and effects",
            "description": "Set the colors and the effects of the bottom LED of the right strip",  # noqa
            "cli": ["--right-strip-bottom-color", "--z7"],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "command": [0x05, 0x00],
            "value_type": "rgbgradient",
            "rgbgradient_header": _RGBGRADIENT_HEADER,
            "led_id": 0x07,
            "default": _DEFAULT_RGBGRADIENT,
        },



    },

    "save_command": {
        "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
        "command": [0x09, 0x00],
    },

}



from . import rival110


# This device looks like a Rival 300 but works like a Rival 110
profile = {

    "name": "SteelSeries Rival 300S",

    "models": [{
        "name": "SteelSeries Rival 300S",
        "vendor_id": 0x1038,
        "product_id": 0x1810,
        "endpoint": 0,
    }],

    "settings": rival110.profile["settings"],

    "save_command": rival110.profile["save_command"],

}



"""
The "rgbgradientv2" type handles RGB color gradients. Simple RGB color can also
be used.

RGB gradient syntax example::

    rgbgradient(duration=1000; colors=0%: #ff0000, 33%: #00ff00, 66%: #0000ff)
    rgbgradient(colors=0%: red, 33%: lime, 66%: blue)

It supports both hexadecimal colors:

* ``#FF0000``
* ``FF0000``
* ``#F00``
* ``F00``

and named colors:

+------------+------------+-----------+-------------+
| ``white``  | ``red``    | ``lime``  | ``blue``    |
+------------+------------+-----------+-------------+
| ``silver`` | ``maroon`` | ``green`` | ``navy``    |
+------------+------------+-----------+-------------+
| ``gray``   | ``yellow`` | ``aqua``  | ``fuchsia`` |
+------------+------------+-----------+-------------+
| ``black``  | ``olive``  | ``teal``  | ``purple``  |
+------------+------------+-----------+-------------+

----

A Python ``dict`` can also be used (Python API only)::

    {
        "duration": 1000,  # ms
        "colors": [
            {"pos": 0, "color": "red"},
            {"pos": 33, "color": "#00FF00"},
            {"pos": 66, "color": (0, 0, 255)},
        ]
    }

.. NOTE::

   A maximum of 14 color stops can be defined in a gradient.


Device Profile
--------------

Example of a rgbgradient value type in a device profile:

::

    profile = {

        # ...

        "settings": {

            "logo_color": {
                "label": "Logo LED colors and effects",
                "description": "Set the colors and the effects of the logo LED",
                "cli": ["-c", "--logo-color"],
                "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
                "command": [0x05, 0x00],
                "value_type": "rgbgradientv2",
                "rgbgradient_header": {
                    "color_field_length": 139,  # Index of length of color field (used for padding)
                    "duration_length": 2,       # Length of the "duration" field (in Bytes)
                    "maxgradient": 14,          # Max numbers of color stop (probably 14)
                },
                "led_id": 0x01,
                "default": "rgbgradient(duration=1000; colors=0%: #ff00e1, 33%: #ffea00, 66%: #00ccff)",
            },

        },

        # ...

    }


CLI
---

Example of CLI option generated with this handler::

   -c LOGO_COLOR, --logo-color LOGO_COLOR
                         Set the colors and the effects of the logo LED (default:
                         rgbgradient(duration=1000; colors=0%: #ff00e1, 33%: #ffea00, 66%: #00ccff))

Example of CLI usage::

    rivalcfg --logo-color="rgbgradient(duration=1000; colors=0%: #ff00e1, 33%: #ffea00, 66%: #00ccff)"
    rivalcfg --logo-color=red
    rivalcfg --logo-color=FF1800


Functions
---------
"""  # noqa


from ..helpers import uint_to_little_endian_bytearray, merge_bytes
from ..helpers import bytes_to_high_low_nibbles, nibbles_to_byte
from ..color_helpers import is_color
from .rgbgradient import _handle_color_tuple, _handle_color_string
from .rgbgradient import _handle_rgbgradient_dict, _handle_rgbgradient_string
from .rgbgradient import is_rgbgradient, CheckGradientAction


# Compatibility with Python 2.7.
# On Python 2 the type is 'unicode'
# On Python 3 the type is 'str'
_unicode_type = type(u"")

_default_duration = 1000


def process_value(setting_info, colors):
    """Called by the :class:`rivalcfg.mouse.Mouse` class when processing a
    "rgbgradientv2" type setting.

    :param dict setting_info: The information dict of the setting from the
                              device profile.
    :param str,tuple,list,dict colors: The color(s).
    :rtype: [int]
    """
    color_field_length = setting_info["rgbgradientv2_header"]["color_field_length"] # noqa
    duration_length = setting_info["rgbgradientv2_header"]["duration_length"]       # noqa
    maxgradient = setting_info["rgbgradientv2_header"]["maxgradient"]               # noqa
    duration = _default_duration
    gradient = []

    # Color tuple
    if type(colors) in (tuple, list):
        gradient = _handle_color_tuple(colors)

    # Simple color string
    elif type(colors) in [str, _unicode_type] and is_color(colors):
        gradient = _handle_color_string(colors)

    # Color gradient as dict
    elif type(colors) is dict:
        duration, gradient = _handle_rgbgradient_dict(colors)

    # Color gradient as string
    elif is_rgbgradient(colors)[0]:
        duration, gradient = _handle_rgbgradient_string(colors)
    else:
        raise ValueError("Not a valid color or rgbgradient %s" % str(colors))

    # -- Check

    if len(gradient) == 0:
        raise ValueError("no color: %s" % str(colors))

    # SSE allowes a maximun of 14 rgbgradient patterns but there is room in the
    # command for up to 16 rgbgradient patterns and it will take 16 arguments.
    gradient_length = len(gradient)
    if gradient_length > maxgradient:
        raise ValueError("a maximum of %i color stops are allowed" %
                         (maxgradient))

    # SSE limits minimum duration depening the the amount of gradient arguments
    minimum_duration = int(gradient_length * 33.3)
    if duration < minimum_duration:
        raise ValueError("a duration of %i or above is need for %i gradient" %
                         (minimum_duration, gradient_length))

    # SSE allows a max duration of 30.00 sec
    if duration > 30000:
        raise ValueError("a maximum duration of 30000ms is allowed")

    # -- Generate header

    start_header = [0x1d, 0x01, 0x02, 0x31, 0x51, 0xff, 0xc8, 0x00]
    #              [0xff, 0x3c, 0x00, 0xff, 0x32, 0xc8, 0xc8, 0x00]
    # [WIP] header command
    header = merge_bytes(setting_info["led_id"], start_header)

    # 8 bytes in a stage, first byte is index, 2nd is padding,
    # 3-5 is signed bytes depecting color increase/decrease, 6 bytes
    # is padding, the 7-8 is time since last stage in ms
    # Process colors and positions in shift array, appending each stage
    # to the previous one

    last_real_pos = gradient[0]["pos"]
    start_color = gradient[0]["color"]
    del gradient[0]

    index = 0
    stage = []
    oldcolor = list(start_color)
    for pos, color in [(item["pos"], item["color"]) for item in gradient]:
        if pos <= last_real_pos:
            raise ValueError("Incorrect order for gradient or duplicate order found please check position order") # noqa
        stage.append(index)  # Stage index number
        stage.append(0)  # Padding
        time = int((duration / 100) * (pos - last_real_pos))
        last_real_pos = pos
        if time == 0:
            raise ValueError("Incompatble timings set, please set different timings") # noqa
        rgb_index = 0
        for rgb in color:
            diff = rgb - oldcolor[rgb_index]
            ramp = int(diff / float(time) * 16)
            oldcolor[rgb_index] = rgb
            stage = merge_bytes(stage, ramp & 255)
            rgb_index = rgb_index + 1
        stage.append(0)  # Padding
        time = uint_to_little_endian_bytearray(time, 2)
        stage = merge_bytes(stage, time)
        index = index + 1

    header = merge_bytes(header, stage)
    # Pad the rest of the command so we can place start color, end suffix and
    # cycle time at the correct offset in the command

    padding = [0] * (color_field_length - len(header))
    header = merge_bytes(header, padding)

    # Split start color into high low nibbles
    split_color = []
    for i in range(len(start_color)):
        high, low = bytes_to_high_low_nibbles(start_color[i])
        left_byte = nibbles_to_byte(low, 0)
        right_byte = nibbles_to_byte(0, high)
        split_color.append(left_byte)
        split_color.append(right_byte)

    end_suffix = [0xff, 0x00]
    # Need to fully test these values
    focal_x = uint_to_little_endian_bytearray(1500, 2)
    focal_y = uint_to_little_endian_bytearray(650, 2)
    end_suffix2 = [0x00, 0x00, 0x00, 0x00, 0x01, 0x00]
    # Amount of colors in gradient (command still work if value is incorrect)
    num_color = uint_to_little_endian_bytearray(gradient_length - 1, 2)
    duration = uint_to_little_endian_bytearray(duration, duration_length)
    suffix = merge_bytes(split_color, end_suffix, focal_x, focal_y,
                         end_suffix2, num_color, duration)

    return merge_bytes(header, suffix)


def add_cli_option(cli_parser, setting_name, setting_info):
    """Add the given "rgbgradientv2" type setting to the given CLI arguments
    parser.

    :param ArgumentParser cli_parser: An :class:`ArgumentParser` instance.
    :param str setting_name: The name of the setting.
    :param dict setting_info: The information dict of the setting from the
                              device profile.
    """
    description = "%s (default: %s)" % (
            setting_info["description"],
            str(setting_info["default"]).replace("%", "%%"),
            )
    cli_parser.add_argument(
            *setting_info["cli"],
            dest=setting_name,
            help=description,
            type=str,
            action=CheckGradientAction,
            metavar=setting_name.upper()
            )



from .. import usbhid

_RGBGRADIENTV2_HEADER = {
    "color_field_length": 139,  # Index of length of color field (used for padding)   # noqa
    "duration_length": 2,       # Length of the "duration" field (in bytes)
    "maxgradient": 14,          # Max numbers of color stop (probably 14)
}

_DEFAULT_RGBGRADIENTV2 = "rgbgradient(duration=1000; colors=0%: #ff00e1, 33%: #ffea00, 66%: #00ccff)"  # noqa

profile = {

    "name": "SteelSeries Rival 700 / SteelSeries Rival 710",

    "models": [{
        "name": "SteelSeries Rival 700",
        "vendor_id": 0x1038,
        "product_id": 0x1700,
        "endpoint": 0,
    }, {
        "name": "SteelSeries Rival 710",
        "vendor_id": 0x1038,
        "product_id": 0x1730,
        "endpoint": 0,
    }],

    "settings": {

        "sensitivity1": {
            "label": "Sensibility preset 1",
            "description": "Set sensitivity preset 1 (DPI)",
            "cli": ["-s", "--sensitivity1"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x03, 0x00, 0x01],
            "command_suffix": [0x00, 0x42],
            "value_type": "range",
            "input_range": [100, 12000, 100],
            "output_range": [0x00, 0x77, 1],
            "default": 800,
        },

        "sensitivity2": {
            "label": "Sensibility preset 2",
            "description": "Set sensitivity preset 2 (DPI)",
            "cli": ["-S", "--sensitivity2"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x03, 0x00, 0x02],
            "command_suffix": [0x00, 0x42],
            "value_type": "range",
            "input_range": [100, 12000, 100],
            "output_range": [0x00, 0x77, 1],
            "default": 1600,
        },

        "polling_rate": {
            "label": "Polling rate",
            "description": "Set polling rate (Hz)",
            "cli": ["-p", "--polling-rate"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x04, 0x00],
            "value_type": "choice",
            "choices": {
                125: 0x04,
                250: 0x03,
                500: 0x02,
                1000: 0x01,
            },
            "default": 1000,
        },

        "logo_color": {
            "label": "Logo LED colors and effects",
            "description": "Set the logo colors and effects",
            "cli": ["-c", "--logo-color"],
            "command": [0x05, 0x00],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "value_type": "rgbgradientv2",
            "rgbgradientv2_header": _RGBGRADIENTV2_HEADER,
            "led_id": 0x0,
            "default": _DEFAULT_RGBGRADIENTV2,
        },

        "wheel_color": {
            "label": "Wheel LED colors and effects",
            "description": "Set the wheel colors and effects",
            "cli": ["-C", "--wheel-color"],
            "command": [0x05, 0x00],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "value_type": "rgbgradientv2",
            "rgbgradientv2_header": _RGBGRADIENTV2_HEADER,
            "led_id": 0x1,
            "default": _DEFAULT_RGBGRADIENTV2,
        },

    },

    "save_command": {
        "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
        "command": [0x09, 0x00],
    },

}



#!/usr/bin/env python
# encoding: UTF-8

import os
from setuptools import setup, find_packages

long_description = ""

try:
    if os.path.isfile("README.rst"):
        long_description = open("README.rst", "r").read()
except Exception as error:
    print("Unable to read the README file: " + str(error))


setup(
    name="rivalcfg",
    version="4.2.0",
    description="Configure SteelSeries gaming mice",
    url="https://github.com/flozz/rivalcfg",
    license="WTFPL",

    long_description=long_description,

    author="Fabien LOISON",

    keywords="steelseries rival sensei mouse",
    platforms=["Linux", "Windows"],

    packages=find_packages(),

    install_requires=[
        "hidapi>=0.7.99.post20"
    ],

    extras_require={
        "dev": [
            "flake8",
            "pytest",
            "Sphinx",
            "sphinx-rtd-theme",
            "nox",
        ],
    },

    entry_points={
        "console_scripts": [
            "rivalcfg = rivalcfg.__main__:main"
        ]
    },

    # cmdclass={"install": install}
)



from .. import usbhid


profile = {

    "name": "SteelSeries Rival 300 / SteelSeries Rival",

    "models": [{
        "name": "SteelSeries Rival",
        "vendor_id": 0x1038,
        "product_id": 0x1384,
        "endpoint": 0,
    }, {
        "name": "SteelSeries Rival Dota 2 Edition",
        "vendor_id": 0x1038,
        "product_id": 0x1392,
        "endpoint": 0,
    }, {
        "name": "SteelSeries Rival 300",
        "vendor_id": 0x1038,
        "product_id": 0x1710,
        "endpoint": 0,
    }, {
        "name": "SteelSeries Rival 300 Fallout 4 Edition",
        "vendor_id": 0x1038,
        "product_id": 0x1712,
        "endpoint": 0,
    }, {
        "name": "SteelSeries Rival 300 Evil Geniuses Edition",
        "vendor_id": 0x1038,
        "product_id": 0x171c,
        "endpoint": 0,
    }, {
        "name": "SteelSeries Rival 300 CS:GO Fade Edition",
        "vendor_id": 0x1038,
        "product_id": 0x1394,
        "endpoint": 0,
    }, {
        "name": "SteelSeries Rival 300 CS:GO Hyper Beast Edition",
        "vendor_id": 0x1038,
        "product_id": 0x171a,
        "endpoint": 0,
    }, {
        "name": "SteelSeries Rival 300 CS:GO Fade Edition (stm32)",
        "vendor_id": 0x1038,
        "product_id": 0x1716,
        "endpoint": 0,
    }, {
        "name": "SteelSeries Rival 300 Acer Predator Edition",
        "vendor_id": 0x1038,
        "product_id": 0x1714,
        "endpoint": 0,
    }, {
        "name": "SteelSeries Rival 300 HP OMEN Edition",
        "vendor_id": 0x1038,
        "product_id": 0x1718,
        "endpoint": 0,
    }],

    "settings": {

        "sensitivity1": {
            "label": "Sensibility preset 1",
            "description": "Set sensitivity preset 1 (DPI)",
            "cli": ["-s", "--sensitivity1"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x03, 0x01],
            "value_type": "range",
            "input_range": [50, 6500, 50],
            "output_range": [0x01, 0x82, 1],
            "default": 800,
        },

        "sensitivity2": {
            "label": "Sensibility preset 2",
            "description": "Set sensitivity preset 2 (DPI)",
            "cli": ["-S", "--sensitivity2"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x03, 0x02],
            "value_type": "range",
            "input_range": [50, 6500, 50],
            "output_range": [0x01, 0x82, 1],
            "default": 1600,
        },

        "polling_rate": {
            "label": "Polling rate",
            "description": "Set polling rate (Hz)",
            "cli": ["-p", "--polling-rate"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x04, 0x00],
            "value_type": "choice",
            "choices": {
                125: 0x04,
                250: 0x03,
                500: 0x02,
                1000: 0x01,
            },
            "default": 1000,
        },

        "logo_color": {
            "label": "Logo LED color",
            "description": "Set the color of the logo LED",
            "cli": ["-c", "--logo-color"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x08, 0x01],
            "value_type": "rgbcolor",
            "default": "#FF1800"
        },

        "wheel_color": {
            "label": "Wheel LED color",
            "description": "Set the color of the wheel LED",
            "cli": ["-C", "--wheel-color"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x08, 0x02],
            "value_type": "rgbcolor",
            "default": "#FF1800"
        },

        "logo_light_effect": {
            "label": "Logo light effect",
            "description": "Set the light effect of the logo",
            "cli": ["-e", "--logo-light-effect"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x07, 0x01],
            "value_type": "choice",
            "choices": {
                "steady": 0x01,
                "breath": 0x03,
                1: 0x01,
                2: 0x02,
                3: 0x03,
                4: 0x04,
            },
            "default": "steady",
        },

        "wheel_light_effect": {
            "label": "Wheel light effect",
            "description": "Set the light effect of the wheel",
            "cli": ["-E", "--wheel-light-effect"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x07, 0x02],
            "value_type": "choice",
            "choices": {
                "steady": 0x01,
                "breath": 0x03,
                1: 0x01,
                2: 0x02,
                3: 0x03,
                4: 0x04,
            },
            "default": "steady",
        },

        "buttons_mapping": {
            "label": "Buttons mapping",
            "description": "Set the mapping of the buttons",
            "cli": ["-b", "--buttons"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x31, 0x00],
            "value_type": "buttons",

            "buttons": {
                "Button1": {"id": 0x01, "offset": 0x00, "default": "button1"},
                "Button2": {"id": 0x02, "offset": 0x05, "default": "button2"},
                "Button3": {"id": 0x03, "offset": 0x0A, "default": "button3"},
                "Button4": {"id": 0x04, "offset": 0x0F, "default": "button4"},
                "Button5": {"id": 0x05, "offset": 0x14, "default": "button5"},
                "Button6": {"id": 0x06, "offset": 0x19, "default": "dpi"},
            },

            "button_field_length": 5,

            "button_disable":     0x00,
            "button_keyboard":    0x51,
            "button_multimedia":  0x61,
            "button_dpi_switch":  0x30,
            "button_scroll_up":   0x31,
            "button_scroll_down": 0x32,

            "default": "buttons(button1=button1; button2=button2; button3=button3; button4=button4; button5=button5; button6=dpi; layout=qwerty)",  # noqa
        },

    },

    "save_command": {
        "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
        "command": [0x09, 0x00],
    },

}



from .. import usbhid


_RGBGRADIENTV2_HEADER = {
    "color_field_length": 139,  # Index of length of color field (used for padding)   # noqa
    "duration_length": 2,       # Length of the "duration" field (in bytes)
    "maxgradient": 14,          # Max numbers of color stop (probably 14)
}

_DEFAULT_RGBGRADIENTV2 = "rgbgradient(duration=1000; colors=0%: #ff00e1, 33%: #ffea00, 66%: #00ccff)"  # noqa


profile = {

    "name": "SteelSeries Rival 500",

    "models": [{
        "name": "SteelSeries Rival 500",
        "vendor_id": 0x1038,
        "product_id": 0x170e,
        "endpoint": 0,
    }],

    "settings": {

        "sensitivity1": {
            "label": "Sensibility preset 1",
            "description": "Set sensitivity preset 1 (DPI)",
            "cli": ["-s", "--sensitivity1"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x03, 0x00, 0x01],
            "command_suffix": [0x00, 0x42],
            "value_type": "range",
            "input_range": [100, 12000, 100],
            "output_range": [0x00, 0x77, 1],
            "default": 800,
        },

        "sensitivity2": {
            "label": "Sensibility preset 2",
            "description": "Set sensitivity preset 2 (DPI)",
            "cli": ["-S", "--sensitivity2"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x03, 0x00, 0x02],
            "command_suffix": [0x00, 0x42],
            "value_type": "range",
            "input_range": [100, 12000, 100],
            "output_range": [0x00, 0x77, 1],
            "default": 1600,
        },

        "polling_rate": {
            "label": "Polling rate",
            "description": "Set polling rate (Hz)",
            "cli": ["-p", "--polling-rate"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x04, 0x00],
            "value_type": "choice",
            "choices": {
                125: 0x04,
                250: 0x03,
                500: 0x02,
                1000: 0x01,
            },
            "default": 1000,
        },

        "logo_color": {
            "label": "Logo LED colors and effects",
            "description": "Set the logo colors and effects",
            "cli": ["-c", "--logo-color"],
            "command": [0x05, 0x00],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "value_type": "rgbgradientv2",
            "rgbgradientv2_header": _RGBGRADIENTV2_HEADER,
            "led_id": 0x0,
            "default": _DEFAULT_RGBGRADIENTV2,
        },

        "wheel_color": {
            "label": "Wheel LED colors and effects",
            "description": "Set the wheel colors and effects",
            "cli": ["-C", "--wheel-color"],
            "command": [0x05, 0x00],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "value_type": "rgbgradientv2",
            "rgbgradientv2_header": _RGBGRADIENTV2_HEADER,
            "led_id": 0x1,
            "default": _DEFAULT_RGBGRADIENTV2,
        },

        "buttons_mapping": {
            "label": "Buttons mapping",
            "description": "Set the mapping of the buttons",
            "cli": ["-b", "--buttons"],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "command": [0x31, 0x00],
            "value_type": "buttons",

            "buttons": {
                "Button1":   {"id": 0x01, "offset": 0x00, "default": "button1"},  # noqa
                "Button2":   {"id": 0x02, "offset": 0x05, "default": "button2"},  # noqa
                "Button3":   {"id": 0x03, "offset": 0x0A, "default": "button3"},  # noqa
                "Button4":   {"id": 0x04, "offset": 0x0F, "default": "button4"},  # noqa
                "Button5":   {"id": 0x05, "offset": 0x14, "default": "button5"},  # noqa
                "Button6":   {"id": 0x06, "offset": 0x19, "default": "button6"},  # noqa
                "Button7":   {"id": 0x07, "offset": 0x1E, "default": "button7"},  # noqa
                "Button9":   {"id": 0x00, "offset": 0x23, "default": "disabled"},  # noqa
                "Button10":  {"id": 0x00, "offset": 0x28, "default": "dpi"},  # noqa
                "Button11":  {"id": 0x00, "offset": 0x2d, "default": "disabled"},  # noqa
                "Button12":  {"id": 0x00, "offset": 0x32, "default": "disabled"},  # noqa
                "TiltLeft":  {"id": 0x33, "offset": 0x37, "default": "TiltLeft"},  # noqa
                "TiltRight": {"id": 0x34, "offset": 0x3c, "default": "TiltRight"},  # noqa
                "Button13":  {"id": 0x00, "offset": 0x41, "default": "disabled"},  # noqa
                "Button8":   {"id": 0x08, "offset": 0x46, "default": "button8"},  # noqa
            },

            "button_field_length": 5,

            "button_disable":     0x00,
            "button_keyboard":    0x51,
            "button_multimedia":  0x61,
            "button_dpi_switch":  0x30,
            "button_scroll_up":   0x31,
            "button_scroll_down": 0x32,

            "default": "default",  # noqa
        },

    },

    "save_command": {
        "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
        "command": [0x09, 0x00],
    },

}



import pytest

from rivalcfg import usbhid


class TestIsDevicePlugged(object):

    def test_a_not_plugged_device(self):
        assert not usbhid.is_device_plugged(0x1038, 0xbaad)

    def test_a_plugged_debug_device(self, monkeypatch):
        monkeypatch.setenv("RIVALCFG_PROFILE", "1038:1702")
        assert usbhid.is_device_plugged(0x1038, 0x1702)


class TestOpenDevice(object):

    def test_a_not_plugged_device(self, monkeypatch):
        monkeypatch.delenv("RIVALCFG_DRY")  # Be sure that dry mode is disabled
        with pytest.raises(usbhid.DeviceNotFound):
            usbhid.open_device(0x1038, 0xbaad, 0x00)

    def test_in_dry_mode(self, monkeypatch):
        monkeypatch.setenv("RIVALCFG_DRY", "1")
        device = usbhid.open_device(0x1038, 0xbaad, 0x00)
        assert hasattr(device, "write")
        assert hasattr(device, "close")
        assert hasattr(device, "send_feature_report")
